/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.conqat.lib.commons.collections.CollectionUtils;
import org.conqat.lib.commons.collections.UnmodifiableList;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.interfaces.IStorageProvider;
import org.fortiss.tooling.kernel.interfaces.ITopLevelElementChangeListener;
import org.fortiss.tooling.kernel.interfaces.ITopLevelElementContext;
import org.fortiss.tooling.kernel.services.IPersistencyService;
import org.fortiss.tooling.kernel.util.ExtensionPointUtils;
import org.fortiss.tooling.kernel.util.LoggingUtils;
import org.osgi.framework.Bundle;

/**
 * This class implements the {@link IPersistencyService} interface. All of its
 * public methods are synchronized.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 92CEC917582981B1AD7A906AF31D5BE0
 */
public class PersistencyService implements IPersistencyService {

	/** The prototype provider extension point ID. */
	private static final String EXTENSION_POINT_NAME = "org.fortiss.tooling.kernel.modelStorageProvider";

	/** The prototype provider configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "modelStorageProvider";

	/** Stores the registered storage providers. */
	private final List<IStorageProvider> storageProviderList = new ArrayList<IStorageProvider>();

	/** Stores the top-level element contexts. */
	private final List<ITopLevelElementContext> contextCache = new LinkedList<ITopLevelElementContext>();

	/** Stores the storage providers and their element caches. */
	private final Map<ITopLevelElementContext, IStorageProvider> storageProviderCache = new HashMap<ITopLevelElementContext, IStorageProvider>();

	/** Stores the top-level element listeners. */
	private final List<ITopLevelElementChangeListener> listeners = new ArrayList<ITopLevelElementChangeListener>();

	/** Constructor. */
	public PersistencyService() {
		setupStorageProviders();
		initializeTopLevelElementContexts();
	}

	/** {@inheritDoc} */
	@Override
	public synchronized UnmodifiableList<ITopLevelElementContext> getTopLevelElementContexts() {
		return CollectionUtils.asUnmodifiable(contextCache);
	}

	/** {@inheritDoc} */
	@Override
	public synchronized boolean isTopLevelElement(EObject element) {
		for (ITopLevelElementContext context : contextCache) {
			if (context.getTopLevelElement() == element) {
				return true;
			}
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public synchronized boolean isDirty() {
		for (ITopLevelElementContext context : contextCache) {
			if (context.isDirty()) {
				return true;
			}
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public synchronized void doSave(IProgressMonitor monitor) {
		for (ITopLevelElementContext context : contextCache) {
			if (context.isDirty()) {
				try {
					context.doSave(monitor);
				} catch (CoreException e) {
					LoggingUtils.error(ToolingKernelActivator.getDefault(),
							"Error during save operation.", e);
				} catch (IOException e) {
					LoggingUtils.error(ToolingKernelActivator.getDefault(),
							"Error during save operation.", e);
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public synchronized void refreshTopLevelElements(IStorageProvider provider) {
		if (provider == null) {
			return;
		}

		List<ITopLevelElementContext> providedElements = provider
				.getTopLevelElementContexts();
		List<ITopLevelElementContext> removedCacheElements = new LinkedList<ITopLevelElementContext>();
		for (ITopLevelElementContext top : contextCache) {
			if (provider == storageProviderCache.get(top)
					&& !providedElements.contains(top)) {
				removedCacheElements.add(top);
			}
		}
		for (ITopLevelElementContext top : providedElements) {
			if (!contextCache.contains(top)) {
				contextCache.add(top);
				storageProviderCache.put(top, provider);
				notifyListenersAboutAdd(top);
			}
		}
		for (ITopLevelElementContext top : removedCacheElements) {
			contextCache.remove(top);
			storageProviderCache.remove(top);
			notifyListenersAboutRemove(top);
		}
	}

	/** {@inheritDoc} */
	@Override
	public synchronized void addTopLevelElementListener(
			ITopLevelElementChangeListener listener) {
		if (!listeners.contains(listener)) {
			listeners.add(listener);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeTopLevelElementListener(
			ITopLevelElementChangeListener listener) {
		listeners.remove(listener);
	}

	/** Notifies listener about top-level element adding. */
	private void notifyListenersAboutAdd(ITopLevelElementContext top) {
		for (ITopLevelElementChangeListener listener : listeners) {
			listener.topLevelElementAdded(top);
		}
	}

	/** Notifies listener about top-level element removal. */
	private synchronized void notifyListenersAboutRemove(
			ITopLevelElementContext top) {
		for (ITopLevelElementChangeListener listener : listeners) {
			listener.topLevelElementRemoved(top);
		}
	}

	/** {@inheritDoc} */
	@Override
	public synchronized ITopLevelElementContext getTopLevelElementContextFor(
			EObject modelElement) {
		while (modelElement != null) {
			for (ITopLevelElementContext context : contextCache) {
				if (context.getTopLevelElement() == modelElement) {
					return context;
				}
			}
			modelElement = modelElement.eContainer();
		}
		return null;
	}

	/** Initializes the storage provider list from plugin extensions. */
	private void setupStorageProviders() {
		for (IConfigurationElement ce : ExtensionPointUtils
				.getConfigurationElements(EXTENSION_POINT_NAME,
						CONFIGURATION_ELEMENT_NAME)) {
			Bundle bundle = ExtensionPointUtils.getBundle(ce);
			try {
				Class<?> handlerClass = ExtensionPointUtils.loadClass(
						ce.getAttribute("provider"), bundle);
				IStorageProvider provider = (IStorageProvider) handlerClass
						.getConstructor().newInstance();
				storageProviderList.add(provider);
			} catch (Exception ex) {
				LoggingUtils.error(ToolingKernelActivator.getDefault(),
						ex.getMessage(), ex);
			}
		}
	}

	/** Initializes the top-level element contexts. */
	private void initializeTopLevelElementContexts() {
		for (IStorageProvider provider : storageProviderList) {
			for (ITopLevelElementContext context : provider
					.getTopLevelElementContexts()) {
				contextCache.add(context);
				storageProviderCache.put(context, provider);
			}
		}
	}
}
